/* geg.c */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <sys/time.h>

#define I_ST 100.0
#define I_INC 0.01

#define POW_SHOW "%2.2f"

#define GENE_MAX 500

#define MAX_INTR 10000

#define E_START 10
#define E_END   500

#define MU_MIN -2.5
#define MU_MAX +2.5

double mu_min = MU_MIN;
double mu_max = MU_MAX;

#define MAX 1000

/*
 * SVID & X/Open
 */
#define	M_E		2.7182818284590452354
#define	M_LOG2E		1.4426950408889634074
#define	M_LOG10E	0.43429448190325182765
#define	M_LN2		0.69314718055994530942
#define	M_LN10		2.30258509299404568402
#define	M_PI		3.14159265358979323846
#define	M_PI_2		1.57079632679489661923
#define	M_PI_4		0.78539816339744830962
#define	M_1_PI		0.31830988618379067154
#define	M_2_PI		0.63661977236758134308
#define	M_2_SQRTPI	1.12837916709551257390
#define	M_SQRT2		1.41421356237309504880
#define	M_SQRT1_2	0.70710678118654752440

double man[MAX];
double con = 1.0L / 1.643933566682L;

#define CELL_MAX 5

typedef struct cell {
  double k;
  double pow;
} CELL, *CELL_PTR;

typedef struct gene {
  CELL cells[CELL_MAX];
  double error;
  int at_top;
} GENE, *GENE_PTR;

/* an array of GENES */
GENE genes[ GENE_MAX ];

int compar ( const void *a, const void *b)
{
  GENE_PTR ga,gb;

  ga = (GENE_PTR)a;
  gb = (GENE_PTR)b;

  if ( ga->error > gb->error ) {
    return 1;
  }
  if ( ga->error == gb->error ) {
    return 0;
  }
  return -1;
}

double get_rand_k ( double min, double max )
{
  double res = drand48();
  double range;

  if ( min < 0.0 ) {
    range = max + fabs(min);
  } else {
    range = max - min;
  }

  res *= range;

  res += min;

  return res;
}

/* randomly populate genes, starting at index 'from' */
void rand_pop ( int from )
{
  int i;
  int j;
  GENE_PTR g;
  double pows[CELL_MAX] = {
    1,2,1,1,1 };

  for ( i = from ; i < GENE_MAX ; i++ ) {
    g = &genes[i];

    g->at_top = 0;
    for ( j = 0 ; j < CELL_MAX-1 ; j++ ) {
      if ( 0 == j ) {
	g->cells[j].k   = 1.0;
      } else {
	g->cells[j].k   = get_rand_k ( -2.0, 2.0 );
      }
      g->cells[j].pow = pows[j]; /* get_rand_k ( -1.5, 1.5 ); */
    }
    g->cells[j].k   = get_rand_k ( -0.000001, 0.000001 );
  }
}

void geg(void)
{
  int i,j;
  double in_res, out_res;
  double sum;
  double deg;
  double ist = I_ST;
  double iinc = I_INC;

  /* initialize array */
  man[0] = 0.0;
  for ( i = 1 ; i < MAX ; i++ ) {
    man[i] = con / ist * ist;

#if 0
    printf("%3d %3.16f\n",
	   i,
	   man[i]);
#endif

    ist += iinc;
  }

}

/* calculate estimated value for this cell at 'loc'
 * using genetic engineering
 */
double est_val ( GENE_PTR g, int loc )
{
  double res = 0.0;
  int i;

  for ( i = 0 ; i < CELL_MAX-1 ; i++ ) {

    res += (g->cells[i].k * pow ( man[loc-i-1], g->cells[i].pow ));

  }
  res += g->cells[i].k;

  return res;
}

/* calculate error for a given gene for all locations */
double est_error ( GENE_PTR g )
{
  int i;
  double error = 0.0;

  for ( i = E_START ; i < E_END /* MAX */ ; i++ ) {
    error += fabs( man[i] - est_val(g,i));
  }

#if 0
  printf("error = %f\n",
	 error);
#endif

  return error;
}

/* calculate error for all genes in all locations */
void est_all ( void )
{
  int i;
  int j;
  int k;

  for ( i = 0 ; i < GENE_MAX ; i++ ) {
    genes[i].error = est_error ( &genes[i] );
  }

  /* sort */
  qsort ( genes, GENE_MAX, sizeof(GENE), compar );

  genes[0].at_top++;

  /* display top 5 */
  for ( i = 0 ; i < 5 ; i++ ) {
    printf("%4d %4d, %12.16f - ",
	   i,
	   genes[i].at_top,
	   genes[i].error);

    for ( k = -1, j = 0 ; j < CELL_MAX-1 ; j++ ) {
      printf("%f*x[%d]^" POW_SHOW " ",
	     genes[i].cells[j].k,
	     k,
	     genes[i].cells[j].pow);
      k -= 1;
    }
    printf("%f\n",
	   genes[i].cells[j].k);
  }
}

/* mutate once */
void mutate_once ( GENE_PTR g )
{
  int j,k;
  double amount;

  /* get cell to mutate */
  /* j = lrand48() % CELL_MAX; */

  for ( j = 0 ; j < CELL_MAX-1 ; j++ ) {

    /* k or pow ??? */
    k = lrand48() % 2;

    /* by this much */
    amount = get_rand_k( mu_min, mu_max );

    /* do the adjustment */

#if 0
    printf("mutate_once: g = 0x%08x, j = %d, k = %d, amount = %f\n",
	   g,
	   j,k,amount);
#endif

    switch ( k )
      {
      case 0:
	g->cells[j].k   += amount;
	break;

      default:
	g->cells[j].pow += amount;
	break;

      }
  }

  /* by this much */
  amount = get_rand_k( mu_min, mu_max );
  g->cells[j].k += amount;

}

/* try to optimize top gene */
GENE_PTR opt_top ( GENE_PTR gi )
{
  GENE_PTR g = gi;
  static GENE gtmp;
  double myinc = 1.0;
  double res;
  int i,j;
  int max = 2000;

  memcpy(&gtmp,g,sizeof(GENE));

  for ( j = 0 ; max > 0 && j < 5 ; j++ ) {

    for ( i = 1 ; max > 0 && i < CELL_MAX ; i++ ) {

      /* k up */
      while ( max-- > 0 ) {
	gtmp.cells[i].k += myinc;
	res = est_error(&gtmp);
	if ( fabs(res - g->error) < 1e-6 || res > g->error ) {
	  gtmp.cells[i].k -= myinc;
	  break;
	}
      }
      /* k down */
      while ( max-- > 0 ) {
	gtmp.cells[i].k -= myinc;
	res = est_error(&gtmp);
	if ( fabs(res - g->error) < 1e-6 || res > g->error ) {
	  gtmp.cells[i].k += myinc;
	  break;
	}
      }
    }

    myinc /= 10.0;
  }

  printf("remaining = %d\n",max);

  if ( gtmp.error > g->error ) {
    return g;
  }

  return &gtmp;
}

/* duplicate best, and mutate them slightly
 *
 * keep top 1/2 performers
 * mutate top 5
 * then get the rest newcomers
 */
int dup_best()
{
  int i = GENE_MAX/2;
  int j;
  int k;
  GENE_PTR g;

  /* optomize top 5 and store */
  for ( j = 0 ; j < 5 ; j++ ) {
    g = opt_top( &genes[j] );
    if ( g != &genes[j] ) {
      memcpy( &genes[i], g, sizeof(GENE));
      i++;
    }
  }

  /* first */
  for ( j = 0 ; j < 10 ; j++ ) {

    /* copy */
    memcpy( &genes[i+j], &genes[0], sizeof(GENE));
    genes[i+j].at_top = 0;

    /* mutate slightly */
    mutate_once( &genes[i+j] );
  }
  i += 10;

  /* second */
  for ( j = 0 ; j < 8 ; j++ ) {

    /* copy */
    memcpy( &genes[i+j], &genes[1], sizeof(GENE));
    genes[i+j].at_top = 0;

    /* mutate slightly */
    mutate_once( &genes[i+j] );
  }
  i += 8;

  /* third */
  for ( j = 0 ; j < 6 ; j++ ) {

    /* copy */
    memcpy( &genes[i+j], &genes[2], sizeof(GENE));
    genes[i+j].at_top = 0;

    /* mutate slightly */
    mutate_once( &genes[i+j] );
  }
  i += 6;

  /* forth */
  for ( j = 0 ; j < 4 ; j++ ) {

    /* copy */
    memcpy( &genes[i+j], &genes[3], sizeof(GENE));
    genes[i+j].at_top = 0;

    /* mutate slightly */
    mutate_once( &genes[i+j] );
  }
  i += 4;

  /* fifth */
  for ( j = 0 ; j < 2 ; j++ ) {

    /* copy */
    memcpy( &genes[i+j], &genes[4], sizeof(GENE));
    genes[i+j].at_top = 0;

    /* mutate slightly */
    mutate_once( &genes[i+j] );
  }
  i += 2;

  /* return where we stopped duplicating */
  return i;
}

int main()
{
  int i,j;
  struct timeval tv;

  gettimeofday(&tv,NULL);
  seed48(&tv);

  geg();

  rand_pop(0);

  for ( i = 1 ; i < MAX_INTR ; i++ ) {

    printf("Interation %d\n",i);

    if ( 0 == (i & 0xff) ) {
      mu_min /= 2.0;
      mu_max /= 2.0;
    }


    est_all();
    j = dup_best();
    rand_pop(j);
  }

  return 0;
}
