/*
 * do manifold processing
 */

#include <stdlib.h>
#include <assert.h>
#include <math.h>
#include <search.h>

#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include "qelem.h"
#include "manifold.h"
#include "points.h"
#include "field.h"

/* define our manifold - points move through this */
MAN man[WIN_SIZE_X][WIN_SIZE_Y];

/* create a new waver */
WAVER_PTR waver_new ( gdouble px, gdouble py, int v, int m )
{
  WAVER_PTR w;

  w = (WAVER_PTR)malloc(sizeof(WAVER));
  assert(w!=NULL);

  w->x  = px;
  w->y  = py;
  w->px = px;
  w->py = py;
  w->v  = v;
  w->m  = m;

  /* returnt to caller */
  return w;
}

/* destroy a waver */
void waver_destroy( WAVER_PTR w )
{
  free(w);
}

#if 0
/* update manifold
 *
 * m     - mass, either +1 or -1
 * [x,y] - location of m
 * [i,j] - my current location
 *
 * Returns: TRUE if in range, else FALSE
 */
int waver_update_cell ( int m, int x, int y, int i, int j )
{
  gdouble dx = 0.0, dy = 0.0;

  if ( i >= 0 && i < WIN_SIZE &&
       j >= 0 && j < WIN_SIZE ) {

    get_one_force ( m, x, y, i, j, &dx, &dy );

    /* update x */
    man[i][j].x += dx;

    /* update y */
    man[i][j].y += dy;

    /* return in range */
    return 1;
  }

  /* return not in range */
  return 0;
}
#endif

#if 0
/* propogate a wave through our manifold
 *
 * Returns: TRUE if we are done, else FALSE
 *
 * Explaination:
 *
 *   A rather strange calculation. We need to construct a box around
 *   our origin point, then walk around the sides of that box
 *   and calculate the value of the wave at that point.
 *
 *   We repeat the calculations, once for ever 'v'.
 *
 */
int waver_propogate ( WAVER_PTR w )
{
  int v = w->v;
  int sx,sy;   /* start here, then stop on the
		* way around when we get here
		*/
  int dx,dy;   /* the number of blocks in the box */
  int i,j;
  int last_cells_changed = 0;

 loop:;

  /* if no more velositys, done */
  if ( v < 1 ) {
    return last_cells_changed;
  }

  last_cells_changed = 0;

  /* update to the next box */
  w->x += 1;
  w->y -= 1;

  /* calculate start/stop point */
  sx = w->x;
  sy = w->y;

#ifdef CP_TRACE
  printf("start/stop point = [%d,%d]\n",
	 sx,sy);
#endif

  /* get the block size */
  dx = 2*(sx - (int)w->px) + 1;
  dy = 2*((int)w->py - sy) + 1;

#ifdef CP_TRACE
  printf("block size = %d, %d\n",
	 dx,dy);

  printf("process top\n");
#endif

  /* process top of box */
  j = sy;
  for ( i = sx ; i > (sx - dx) ; i-- ) {
    /* update   the [i,j] cell */
    /* to pt to the [x,y] cell */
    last_cells_changed += waver_update_cell ( w->m, (int)w->px, (int)w->py, i, j );
  }

#ifdef CP_TRACE
  printf("process bottom\n");
#endif

  /* process bottom of box */
  j = sy + dy - 1;
  for ( i = sx ; i > (sx - dx) ; i-- ) {
    /* update   the [i,j] cell */
    /* to pt to the [x,y] cell */
    last_cells_changed += waver_update_cell ( w->m, (int)w->px, (int)w->py, i, j );
  }

#ifdef CP_TRACE
  printf("process left\n");
#endif

  /* process left side of box */
  i = sx - dx + 1;
  for ( j = sy + 1 ; j < (sy + dy - 1) ; j++ ) {
    /* update   the [i,j] cell */
    /* to pt to the [x,y] cell */
    last_cells_changed += waver_update_cell ( w->m, (int)w->px, (int)w->py, i, j );
  }

#ifdef CP_TRACE
  printf("process right\n");
#endif

  /* process right side of box */
  i = sx;
  for ( j = sy + 1 ; j < (sy + dy - 1) ; j++ ) {
    /* update   the [i,j] cell */
    /* to pt to the [x,y] cell */
    last_cells_changed += waver_update_cell ( w->m, (int)w->px, (int)w->py, i, j );
  }

  /* one velosity consumed */
  v--;

  /* did we change anything ??? */
  if ( 0 == last_cells_changed ) {
    /* no, so done */
    return 0;
  }

  /* yes, continue */
  goto loop;
}
#endif

/* zero the manifold */
void zero_manifold ( void )
{
  int i;
  int j;

  for ( i = 0 ; i < WIN_SIZE_X ; i++ ) {
    for ( j = 0 ; j < WIN_SIZE_Y ; j++ ) {
      man[i][j].x = 0.0;
      man[i][j].y = 0.0;
    }
  }
}

#if 0
/* initialize manifold for a point */
static void init_manifold ( int x, int y )
{
  int i;
  int j;

  for ( i = 0 ; i < WIN_SIZE ; i++ ) {
    for ( j = 0 ; j < WIN_SIZE ; j++ ) {
      int dx;
      int dy;
      gdouble f;

      /* for the point, skip the calculation */
      if ( x == i && y == j ) {
	continue;
      }

      dx = i - x;
      if ( dx == 0 ) {
	f = 0.0;
      } else {
	f = 1.0 / (double)( dx*dx );
      }
      man[i][j].x += (i > x ? -1.0 * f : f);

      dy = j - y;
      if ( dy == 0 ) {
	f = 0.0;
      } else {
	f = 1.0 / (double)( dy*dy );
      }
      man[i][j].y += (j > y ? -1.0 * f : f);
    }
  }
}
#endif

#if 0
/* initialize manifold for two points */
void init_two_manifolds ( void )
{
  zero_manifold ( );

#if 0
  init_manifold ( p1.x, p1.y );
  init_manifold ( p2.x, p2.y );
#endif
}
#endif
