/* myidle.c */

#include <stdio.h>
#include <unistd.h>
#include <math.h>
#include <search.h>

#include <gtk/gtk.h>
#include <gdk/gdk.h>

#include "qelem.h"
#include "points.h"
#include "manifold.h"
#include "myidle.h"
#include "scribble-simple.h"
#include "field.h"
#include "sym.h"

/* overall trace flag */
int trace_flag = 0;

/* state engine state */
gint idle_state = 0;

/* current simulation time */
gint current_time_cnt = 0;
double current_time = 0.0;

/* number of seconds for each baby step in simulation */
#define SYM_INC 0.001
double sym_inc = SYM_INC;

/* show aspect */
extern void show_aspect();

/* run helper */
static int my_simulate ( void )
{
  static int i = 1000000000; /* max interation */
  char wbuf[256];
  static int sym_stopped = 0;
  int j;
  static int px_ok_flag = 0;
  static int px_ok[POINT_MAX];
  static int opt_cnt = 0;

  /* initialize point flags */
  if ( 0 == px_ok_flag ) {
    for ( j = 0 ; j < POINT_MAX ; j++ ) {
      px_ok[j] = 1;
    }
    px_ok_flag = 1;
  }

  /* get rid of new flag */
  no_new ( );

  if ( i-- > 0 ) {
    for ( j = 0 ; j < pts_idx ; j++ ) {

      /* see if we want to move this point */
      if ( pts[j].flag & 0x1 ) {
	/* nope, just let it hang out */
	continue;
      }

      /* move px */
      if ( px_ok[j] && !mover_once ( &pts[j], sym_inc ) ) {
	sprintf(wbuf,"Simulation Stopped (p%d) at time = %f\n",
		j,current_time);
	gtk_text_insert (GTK_TEXT(text), fixed_font, NULL, NULL,wbuf, -1 );
	sym_stopped++;
	px_ok[j] = 0;
	if ( pts_idx == sym_stopped ) {
	  return 0;
	}
      }  /* px_ok */
    } /* j */

    /* propogate any fields */
    tick_all_fields ( sym_inc );

    /* show where points have gone */
    draw_all_points();

    /* keep track of simulation time */
    if ( 0 == (++current_time_cnt & 0xfff) ) {

#if 1
      /* optomize number of fields now and again */
      opt_cnt += opt_fields();
#endif

      current_time_cnt = 0;
      sprintf(wbuf,"time = %f, q = %d, opt = %d\n",
	      current_time,
	      num_queued_fields,
	      opt_cnt);
      gtk_text_insert ( GTK_TEXT(text) , fixed_font, NULL, NULL,
		       wbuf, -1
			 );

      if ( 1 & trace_flag ) {
	/* show points velosity */
	show_points_vel ( );
      }

      opt_cnt = 0;
    }

    /* update simulation time */
    current_time += sym_inc;

    /* return to caller */
    return 1;
  }

  /* done, out of ticks */
  return 0;
}

/* create some fields from the points, advance to next state */
static void my_init_fields ( void )
{
  int i;

  /* onto next state */
  idle_state++;

  for ( i = 0 ; i < pts_idx ; i++ ) {

    /* uniforce ??? */
    if ( !(POINT_FLAG_UNIFORCE & pts[i].flag) ) {
      /* no */
      new_field ( pts[i].x,
		  pts[i].y,
		  pts[i].mass,
		  i+1,
		  pts[i].flag,
		  sqrt(WIN_SIZE_X * WIN_SIZE_Y)
		  );
    } else {
      /* yes */
      new_field ( pts[i].x,
		  pts[i].y,
		  pts[i].mass,
		  i+1,
		  pts[i].flag,
		  30            /* use radius 30 for uniforce */
		  );
    }
  }

#ifdef CP_TRACE
  printf("gravity has started to propogate from points\n");
#endif
}

/* initialize */
static void my_init ( void )
{
  /* advance to next state */
  idle_state++;

  /* get initial conditions from 'sym.dat' */
  sym_init ();

  /* initialize manifold */
  zero_manifold();

  /* draw the points */
  draw_all_points();
}

/* draw some grid lines */
void draw_grid ( )
{
  int x,y;

#define LININC 8

  for ( y = 0 ; y < 640 ; y += LININC ) {
    gdk_draw_line ( pixmap,
		    drawing_area->style->black_gc,
		    0,y,
		    640,y );
  }

#if 1
  for ( x = 0 ; x < 640 ; x += LININC ) {
    gdk_draw_line ( pixmap,
		    drawing_area->style->black_gc,
		    x,0,
		    x, 640);
  }
#endif

}

/* draw circs */
void draw_circs ( int x, int y, int r )
{
  int left_x = x - r;
  int top_y  = y - r;
  int wid = 2 * r;
  int hid = 2 * r;

  draw_grid();

  gdk_draw_arc (
		pixmap,
		drawing_area->style->black_gc,
		FALSE,
		left_x, top_y,
		wid, hid,
		0,
		64 * 360 );

#if 0  
  gtk_widget_queue_draw_area (drawing_area, 
			      0,0,
			      400,400);
#endif

}

/* draw elipse */
void draw_elipse ( int x, int y, int r, double div )
{
  int left_x = x - r;
  int top_y  = y - ((double)r)*div;
  int wid = 2 * r;
  int hid = 2 * ((double)r)*div;

  draw_grid();

  gdk_draw_arc (
		pixmap,
		drawing_area->style->black_gc,
		FALSE,
		left_x, top_y,
		wid, hid,
		0,
		64 * 360 );

#if 0  
  gtk_widget_queue_draw_area (drawing_area, 
			      0,0,
			      400,400);
#endif

}

/* undraw elipse */
void undraw_elipse ()
{
  gdk_draw_rectangle (pixmap,
		      drawing_area->style->white_gc,
		      TRUE,
		      0, 0,
		      640,480);
}

int run_flag       = 0;
double cur_div     = 0.999;    /* INITIAL ASPECT */
double cur_div_inc = -0.001;

/* inc */
void do_inc()
{
  cur_div += cur_div_inc;
}

/* dec */
void do_dec()
{
  cur_div -= cur_div_inc;
}

/* draw an arc */
int draw_arc ( )
{
  int i;

  //g_printf("drawing arcs, %lf\n",cur_div);

  if ( run_flag ) {
    show_aspect();
    usleep(100000);
  }

  undraw_elipse ();

  for ( i = 0 ; i < 40 ; i++ ) {
    draw_elipse   ( 320, 240, 50 + 5*i, cur_div );
  }

  if ( run_flag ) do_inc();

  if ( cur_div < 0.1 ) {
    return 0;
  }

#if 0
  for ( i = 0 ; i < 40 ; i++ ) {
    draw_circs ( 300, 300, 50 + 5*i );
  }
#endif

#if 0
  for ( i = 0 ; i < 10 ; i++ ) {
    draw_circs ( 155, 200, 20 + 5*i );
  }
#endif

#if 0
  for ( i = 0 ; i < 3 ; i++ ) {
    draw_circs ( 160, 150, 10 + 5*i*i );
  }
#endif

  gtk_widget_queue_draw_area (drawing_area, 
			      0,0,
			      WIN_SIZE_X, WIN_SIZE_Y);

  return run_flag ? 1 : 0;
}

gint nap ()
{
  return 0;
}

/* Use gtk idle time for simulation,
 * return FALSE if we're done else
 * TRUE
 */
gint my_idle ( gpointer data )
{
  gint sts = 1;

  /* off to appropriate state */
  switch ( idle_state )
    {
    case 0:
      //my_init();
      sts = draw_arc();
      break;

    case 1:
      //my_init_fields();
      sts = draw_arc();
      break;

    case 2:
      sts = my_simulate();
      break;
    }

  /* return to caller */
  return sts;
}
